
//	FILE:	UTIL.C

#include <windows.h>
#include "wterm.h"

BOOL	Initialized_buffer	= FALSE;

// Function: Get_Screen_Pointer
// 			Return a pointer to the specified line on the display.

char *Get_Screen_Pointer(short Index)
{
	short Line_Pointer;

	// get the pointer to the first line of display

	Line_Pointer = Last_Line + 1;
	if(Line_Pointer == MaxLines)
		Line_Pointer = 0;

	Line_Pointer = Line_Pointer + Index;
	if(Line_Pointer > MaxLine)
		Line_Pointer = Line_Pointer - MaxLines;

	return(Screen_Buffer[Line_Pointer]);
}

// Function: Set_Device_Context

void Set_Device_Context(HWND hWnd, HDC hDC)
{
	RECT rClRect;

	GetClientRect(hWnd, &rClRect);
	SetViewportOrg(hDC, 0, rClRect.bottom - ((MaxLines) * Char_Height));
	SetMapMode(hDC, MM_ANISOTROPIC);
	SetWindowExt(hDC, 1, 1);	// one unit on screen is the size of char.

	// Set the last line in the buffer at the bottom of the window. 

	SetViewportExt(hDC, Char_Width, Char_Height);

	// Set the background mode and select the font.

	SetBkMode(hDC, OPAQUE);
	Old_Font = SelectObject(hDC, GetStockObject(My_Font));
}

// Function: Paint_Window
// 			Draws text in the window.

void Paint_Window(HWND hWnd )	
{
	PAINTSTRUCT	ps;
	HDC			hDC;
	RECT		rcUpdate,	rcClient;

	register	int	i;

	char		*Line_Ptr;
	int			Vertical_Beg, Vertical_End,
				Horizontal_Beg, Horizontal_End;

	hDC = BeginPaint( hWnd, (LPPAINTSTRUCT)&ps );
	Set_Device_Context(hWnd,hDC);

	rcUpdate = ps.rcPaint;
	DPtoLP(hDC, (POINT *) &rcUpdate, 2);

	// get first and last lines to be updated

	Vertical_Beg = max (0, rcUpdate.top);
	Vertical_End = min (MaxLines, rcUpdate.bottom);

	// get the first and last columns

	Horizontal_Beg = max (0, rcUpdate.left);
	Horizontal_End = min (NumColumn, rcUpdate.right);

	// display the lines that need to be drawn

	for(i=Vertical_Beg; i<Vertical_End; i++)
	{
		Line_Ptr = Get_Screen_Pointer(i) + Horizontal_Beg;
		TextOut(	hDC,
					Horizontal_Beg,
					i,
					Line_Ptr,
					strlen(Line_Ptr));
	}
	SelectObject(hDC, Old_Font);		// get old font
	EndPaint( hWnd, (LPPAINTSTRUCT)&ps );
}


// Function: Update_Window
// 			Scroll the window by the number of lines received,

void Update_Window(HWND hWnd, HDC hDC, int Num_Lines)
{
	RECT rcRect;

	HideCaret (hWnd);
	if(Num_Lines > 0)		// then scroll screen by Num_lines
	{
		GetClientRect(hWnd, &rcRect);
		rcRect.bottom -= Char_Height;
	
		ScrollWindow(hWnd, 0, -(Char_Height * Num_Lines), &rcRect, NULL);
	}
	UpdateWindow(hWnd);
	SetCaretPos ((Char_Width * Cusor_Pos), (Char_Height * (MaxLine-1)));
	ShowCaret (hWnd);
}


// Function: Process_Text_Buffer

void Process_Text_Buffer(HWND hWnd, WORD Num_Chars, LPSTR lParam)
{
	RECT	rClRect, rcInvalid;
	HDC		hDC;

	short	i, j;
	char	*Buffer_Ptr;
	char	*Line_Ptr;
	int		Num_Lines = 0;						// number of lines to scroll

	hDC		= GetDC(hWnd);

	Set_Device_Context(hWnd, hDC);
	Buffer_Ptr = Screen_Buffer[Last_Line];		// pointer to current line

	// scan the text, handle any special characters, and display the rest.

	for(i=0; i < Num_Chars; i++){
	switch(lParam[i])
	{
		case '\r': 					// return

			Cusor_Pos = 0;			// move to beginning of the line
			break;

		case '\n': 					// new line

			++Num_Lines;			// increment lines to scroll
			++Last_Line;
			Cusor_Pos = 0;			// current position is 0

			if(Last_Line > MaxLine)
				Last_Line = 0;

			// clear the last line on screen

			Buffer_Ptr = Screen_Buffer[Last_Line];
			for(j=0; j<NumColumn; j++)
				Buffer_Ptr[j] = '\0';

			// invalidate the area so that it gets redrawn

			rcInvalid.top		= MaxLine;
			rcInvalid.bottom	= MaxLine + 1;
			rcInvalid.left		= Cusor_Pos;
			rcInvalid.right		= Cusor_Pos + NumColumn;
			LPtoDP(hDC, (POINT *) &rcInvalid,  2);

			InvalidateRect(hWnd, &rcInvalid, TRUE);

			break;

		case '\b':					// backspace

			if(Cusor_Pos > 0)
			{
				--Cusor_Pos;

				rcInvalid.top		= MaxLine;
				rcInvalid.bottom	= MaxLine + 1;
				rcInvalid.left		= Cusor_Pos;
				rcInvalid.right		= Cusor_Pos + 1;
				LPtoDP(hDC, (POINT *) &rcInvalid,  2);

				// invalidate the area so that it gets redrawn

				InvalidateRect(hWnd, &rcInvalid, TRUE);
			}
			break;

		case '\t':
			break;

		default:									// all displayable characters

			// Filter out all characters not supported by windows
			if((lParam[i] < 32)
				|| ((lParam[i] > 126) && (lParam[i] < 145))
				|| ((lParam[i] > 146) && (lParam[i] < 160)))
				break;

			if(Cusor_Pos < NumColumn)
			{
				Buffer_Ptr[Cusor_Pos] = lParam[i]; // add char to screen buffer

				// calculate area to invalidate
				rcInvalid.top = MaxLine; rcInvalid.bottom = MaxLine + 1;
				rcInvalid.left = Cusor_Pos;
				++Cusor_Pos;
				rcInvalid.right = Cusor_Pos;

				if(Num_Lines == 0)				// if this is the last line
				{
					LPtoDP(hDC, (POINT *) &rcInvalid, 2);
					// invalidate the area so that it gets redrawn
					InvalidateRect(hWnd, &rcInvalid, FALSE);
				}
			}
			break;
	}

	if(Num_Lines > 0)
	{
		Update_Window(hWnd, hDC, Num_Lines);
		Num_Lines = 0;
	}
	}

	Update_Window(hWnd, hDC, Num_Lines);
	SelectObject(hDC, Old_Font);		// get old font
	ReleaseDC(hWnd, hDC);
}

// Function: Init_Buffer
//			Clear the display buffer and initialize variables.

void Init_Buffer(HWND hWnd)
{
	int			i, j;
	HDC			hDC;
	TEXTMETRIC	Metrics;

	// Initialize screen buffer to nulls 

	for(i=0; i<MaxLines; i++)
		for(j=0; j<81; j++)
			Screen_Buffer[i][j] = '\0';

	First_Line	= 0;
	Last_Line	= MaxLine;
	Cusor_Pos	= 0;

	// get the text metrics for the font we are using

	hDC			= GetDC(hWnd);
	Old_Font	= SelectObject(hDC, GetStockObject(My_Font));

	GetTextMetrics(hDC, &Metrics);
	SelectObject(hDC, Old_Font);		// get old font
	ReleaseDC(hWnd, hDC);

	// get the height and width of the font 

	Char_Width	= Metrics.tmMaxCharWidth;
	Char_Height	= Metrics.tmHeight + Metrics.tmExternalLeading;

	// get the background and forground colors

	Background	= GetSysColor(COLOR_WINDOW);		// background color
	Text_Color	= GetSysColor(COLOR_WINDOWTEXT);	// text color

	Initialized_buffer = TRUE;
}

// Function: Wterm_Message_proc(HWND, unsigned, WORD, LONG)
// 			Process messages.

long FAR PASCAL Wterm_Message_proc( hWnd, message, wParam, lParam )

HWND		hWnd;
unsigned	message;
WORD		wParam;
LONG		lParam;
{
	PAINTSTRUCT		ps;
	HDC				hDC;
	LPPOINT			ptMinMaxInfo;

	switch (message)
	{
		case WM_CREATE:

			Init_Buffer(hWnd);			// Initialize the display buffer
			break;

		case WM_SYSCOLORCHANGE:

			// if the colors have been changed in the control panel,
			// we need to change also.

			Background = GetSysColor(COLOR_WINDOW);			// background color
			Text_Color = GetSysColor(COLOR_WINDOWTEXT);		// text color
			return DefWindowProc( hWnd, message, wParam, lParam );
			break;

		case WM_GETMINMAXINFO:

			if(!Initialized_buffer)
				Init_Buffer(hWnd);

			// constrain the sizing of the window to 80 by 25 characters.

			ptMinMaxInfo = (LPPOINT) lParam;

			ptMinMaxInfo[1].x = Char_Width * NumColumn
					+ 2 * GetSystemMetrics(SM_CXFRAME);

			ptMinMaxInfo[1].y = Char_Height * (MaxLines+1)
					+ 2 * GetSystemMetrics(SM_CYFRAME);

			ptMinMaxInfo[4].x = Char_Width * NumColumn
					+ 2 * GetSystemMetrics(SM_CXFRAME);

			ptMinMaxInfo[4].y = Char_Height * (MaxLines+1)
					+ 2 * GetSystemMetrics(SM_CYFRAME);
			break;

		case WM_PAINT:

			Paint_Window(hWnd);
			break;

		case WM_DESTROY:

			if(Quit_flg)
				PostQuitMessage(0);
			break;

		case WM_CLOSE:

			// then close the current window
		default:

			return DefWindowProc( hWnd, message, wParam, lParam );
			break;
	}
	return(0L);
}
