#include <windows.h>
#include <stdlib.h>
#include <string.h>
#include "wterm.h"

/*------------------------------------------------------------------------*
 *	WinMain                                                               *
 *	                                                                      *
 *	input:		HANDLE; to the instance                                   *
 *	            HANDLE; to the previous instance                          *
 *	            LPSTR; pointer to the commad line parameters              *
 *	            int; specifies the type of window to open - full or icon  *
 *	output:     none                                                      *
 *	returns:    none                                                      *
 *	globals:    hinst                                                     *
 *	            pBuffer                                                   *
 *	            in_buf                                                    *
 *	            Num_Bytes                                                 *
 *	calls:      Windows                                                   *
 *	abstract:   This is the main Windows function that is called on entry *
 *	            and then processes all Windows messages.                  *
 *	            It first initialises the instance (global data, etc) then *
 *	            enters a doforever loop waiting for messages. Actually it *
 *	            PEEKs for messages which temporarily yields control to    *
 *	            Windows (to let other Windows Apps run). When control is  *
 *	            returned we see if there were any messages. If not then   *
 *	            we check if there are any characters waiting to be        *
 *	            displayed.                                                *
 *	                                                                      *
 *------------------------------------------------------------------------*/



int PASCAL WinMain (HANDLE hInstance,		// Current instance
					HANDLE hPrevInstance,	// Previous instance
					LPSTR  lpCmdLine,		// Command line parameters
					int    nCmdShow)		// Show-window type: open or icon

{
	MSG			msg;
	WNDCLASS	Win_Class;
	HWND		hWnd;				// Main window handle
	BOOL		Message_available;
	unsigned	int Num_Bytes;

    if (!hPrevInstance)   // first instance 
    {
        if (!InitApplication (hInstance))			// init Appl level data
            return (FALSE);   // return false if window registration failed
		hWnd = InitInstance (hInstance, nCmdShow);	// init instance data
    	if (!hWnd)          // if null means create window failed
        	return (FALSE);

		hInst = hInstance;			//  save token for this instance
    }
	else                  // second instance, application already exists
	{
		MessageBeep(2); // since iRMX only has one CONSOLE
		return (FALSE);
	}

	// Process windows messages
	// Loop until a WM_QUIT message is received
	//		If there is a message available, process it.
	//		Otherwise check the iRMX output buffer and handle
	//		any available characters.

	while (TRUE)
		{
			Message_available = PeekMessage(&msg, NULL, 0, 0, PM_REMOVE);

			if (Message_available)
			{
				if (msg.message == WM_QUIT)
					break;

				TranslateMessage ((LPMSG)&msg);
				DispatchMessage ((LPMSG)&msg);
			}
			else
			{
				if ((Connected) && (in_buf->count > 0) && (!Stop_output))
				{
					// copy data from iRMX to local buffer

					for (Num_Bytes=0; Num_Bytes < in_buf->count; Num_Bytes++)
						MsgBuff[Num_Bytes] = in_buf->data_byte[Num_Bytes];

					in_buf->count = 0;	// tell RMX we read the data
								// then display text
					Process_Text_Buffer (hWnd, Num_Bytes, (LPSTR)MsgBuff);
				}
			}
		}			// end forever loop

	free (pBuffer);
	return (msg.wParam);	   // Returns the value from PostQuitMessage
}

/*------------------------------------------------------------------------*
 *	InitApplication                                                       *
 *	                                                                      *
 *	input:		HANDLE; to the instance                                   *
 *	output:     none                                                      *
 *	returns:    TRUE if windows class is registered                       *
 *	globals:    none                                                      *
 *	calls:      Windows                                                   *
 *	abstract:   Registers the Windows Class for this Application.         *
 *	                                                                      *
 *------------------------------------------------------------------------*/

BOOL InitApplication (HANDLE hInstance)

{
	WNDCLASS	Win_Class;

	// Set up window class structure with parameters that describe the
	// main Wterm window.

	Win_Class.style			= CS_HREDRAW | CS_VREDRAW; // Class style(s).
	Win_Class.cbClsExtra	= 0;			// No per-class extra data
	Win_Class.cbWndExtra	= 0;			// No per-window extra data
	Win_Class.hInstance		= hInstance;	// Appln that owns the class

	Win_Class.hIcon			= LoadIcon (NULL, IDI_APPLICATION);
	Win_Class.hCursor		= LoadCursor (NULL, IDC_ARROW);
	Win_Class.hbrBackground	= GetStockObject (WHITE_BRUSH); 

	Win_Class.lpfnWndProc	= MainWndProc;	// Function to receive messages
	Win_Class.lpszMenuName	= "Wterm_Menu";	// Name of menu in .RC file
	Win_Class.lpszClassName	= "WtermWClass";// Class name

	// Register the window class

	return (RegisterClass (&Win_Class));
}

/*------------------------------------------------------------------------*
 *	InitInstance                                                          *
 *	                                                                      *
 *	input:		HANDLE; to the instance                                   *
 *	            int; nCmdShow windows parameter                           *
 *	output:     none                                                      *
 *	returns:    HWND handle if the window is created and init OK          *
 *	globals:    xCaret                                                    *
 *              yCaret		                                              *
 *              Connected                                                 *
 *              iLineLength                                               *
 *              iNumberLines                                              *
 *              frg_color	                                              *
 *              bkg_color	                                              *
 *              in_buf   	                                              *
 *              out_buf                                                   *
 *	calls:      Windows                                                   *
 *	abstract:   Creates the window, inits the globals, allocates memory   *
 *	            and shows the window.                                     *
 *	                                                                      *
 *------------------------------------------------------------------------*/

HWND InitInstance (HANDLE hInstance, int nCmdShow)

{
	HWND	hWnd;

	// Create a main window for this application instance

	hWnd = CreateWindow(
			"WtermWClass",						// See RegisterClass() call
			"WTERM",							// Text for window title bar
			WS_OVERLAPPEDWINDOW,				// Window style
			CW_USEDEFAULT,						// initial x  pos
			CW_USEDEFAULT,						// initial y  pos
			GetSystemMetrics(SM_CXSCREEN) / 2,	// initial window width
			GetSystemMetrics(SM_CYSCREEN) / 2,	// initial window height
			NULL,								// Overlapped windows have no parent
			NULL,								// Use the window class menu
			hInstance,							// This instance owns this window
			NULL								// Pointer not needed
		);

	// If CreateWindow call failed - just say NO

	if (!hWnd)
		return ((HWND)NULL);

	// set icon for WTERM

	SetClassWord (hWnd,
				 GCW_HICON,
				 LoadIcon (hInstance, MAKEINTRESOURCE(WTERMICON)));

	// initialize global variables

	xCaret		= 0;
	yCaret		= 0;
	iLineLength	= MAX_COLUMN;
	iNumberLines= MAX_ROW;
	frg_color	= 0;	// Black
	bkg_color	= 0xf;	// White

  	if (RQEGetRmxStatus() != EOK)
	{
		MessageBox(hWnd,
			"iRMX Operating System not present"
			" for Wterm.", "iRMX Terminal",
			MB_ICONEXCLAMATION | MB_OK);
		return (FALSE);
	}
  
	if ((pBuffer = malloc (iLineLength * iNumberLines)) == NULL)
	{
		MessageBox(hWnd,
			"Cannot allocate enought memory"
			" for WTERM.", "",
			MB_ICONSTOP | MB_OK);
		return ((HWND)NULL);
	}
	else
		memset (pBuffer, ' ', iLineLength * iNumberLines);

	// Initialize input/output buffer @ iRMX comm. area

	//	Note: We are using 32 bytes of DOS memory at the end
	//		of the BIOS communication area, starting at 4E0H

	in_buf->count	= 0;

	out_buf->flag_byte	= 0;
	out_buf->data_byte	= 0;

	// Make the window visible and update client area

	ShowWindow (hWnd, nCmdShow);
	UpdateWindow (hWnd);			// Sends WM_PAINT message
	return (hWnd);
}

/*------------------------------------------------------------------------*
 *	InitWindow                                                            *
 *	                                                                      *
 *	input:		HWND; Window handle                                       *
 *	output:     none                                                      *
 *	returns:    none                                                      *
 *	globals:    none                                                      *
 *	calls:      GetDC                                                     *
 *	            SelectObject                                              *
 *	            GetTextMetrics                                            *
 *	            ReleaseDC                                                 *
 *	abstract:   Sets up the Width and Height of the CHARs                 *
 *	                                                                      *
 *------------------------------------------------------------------------*/

void InitWindow (HWND hWnd)

{
	TEXTMETRIC	tm;
	HDC			hdc;
	
	hdc = GetDC(hWnd);

	SelectObject (hdc, GetStockObject(SYSTEM_FIXED_FONT));
	GetTextMetrics (hdc, &tm);
	iCharWidth  = tm.tmAveCharWidth;
	iCharHeight = tm.tmHeight;
	HideCaret (hWnd);

	ReleaseDC (hWnd, hdc);
	WinCreated = TRUE;
}

/*------------------------------------------------------------------------*
 *	MainWndProc                                                           *
 *	                                                                      *
 *	input:		HWND; Window handle                                       *
 *	            unsigned; Message type                                    *
 *	            WORD; wParam                                              *
 *	            LONG; lParam                                              *
 *	output:     none                                                      *
 *	returns:    none                                                      *
 *	globals:    iCharWidth                                                *
 *	            iCharHeight                                               *
 *	            xCaret                                                    *
 *	            yCaret                                                    *
 *	            Stop_Output                                               *
 *	            iLineLength                                               *
 *	            iNumberLines                                              *
 *	calls:      WinCreated                                                *
 *	            InitWindow                                                *
 *	            Paint_Window                                              *
 *	            Windows                                                   *
 *	abstract:   This is the main Windows message processing function for  *
 *	            the Application. It does pretty much standard processing  *
 *	            for most of the messages; the special hangling is confined*
 *	            to the Keyboard character handling. Every time a key is   *
 *	            hit, it is put in the shared data area and a message is   *
 *	            sent to iRMX.                                             *
 *	                                                                      *
 *------------------------------------------------------------------------*/

long FAR PASCAL MainWndProc (HWND hWnd,
							 unsigned message,
							 WORD wParam,
							 LONG lParam)

{
	FARPROC		lpProcAbout;		// pointer to the "About" function
	LPPOINT		ptMinMaxInfo;
	int			MaxWindowHeight,
				MaxWindowWidth;

	switch (message)
	{
		case WM_SIZE:
			break;

		case WM_CREATE:

			if (!WinCreated)
				InitWindow (hWnd);
			break;

		case WM_GETMINMAXINFO:

			if (!WinCreated)
				InitWindow (hWnd);
				
			// constrain the sizing of the window to 80 by 25 characters.

			MaxWindowHeight	= iCharHeight * MAX_ROW
								+ (2 * GetSystemMetrics (SM_CYFRAME))
								+ GetSystemMetrics (SM_CYCAPTION)
								+ GetSystemMetrics (SM_CYMENU);
			MaxWindowWidth	= iCharWidth * MAX_COLUMN
								+ (2 * GetSystemMetrics (SM_CXFRAME));
			
			ptMinMaxInfo = (LPPOINT) lParam;

			/*
			** [1] defines the maximised size that the window can be
			**     resized to using the MAXIMIZE button; in our
			**     case, it's 80 x 25 characters
			*/
			ptMinMaxInfo[1].x = MaxWindowWidth;
			ptMinMaxInfo[1].y = MaxWindowHeight;

			/*
			** [2] defines the Upper Left hand corner position
			**     of the maximised window; in our case it's
			**     left upto Windows
			*/

			/*
			** [3] defines the minimum size that the window can be
			**     resized to using the borders; in our case it's
			**     a quarter of the screen width and the height is
			**     FIXED to ensure 25 ROWs are always displayed.
			*/
			ptMinMaxInfo[3].x = (GetSystemMetrics (SM_CXSCREEN) / 4);
			ptMinMaxInfo[3].y = MaxWindowHeight;

			/*
			** [4] defines the maximised size that the window can be
			**     resized to using the borders; in our case it's
			**     the same as using the MAXIMIZE button
			*/
			ptMinMaxInfo[4].x = MaxWindowWidth;
			ptMinMaxInfo[4].y = MaxWindowHeight;

			break;

		case WM_SETFOCUS:

			CreateCaret (hWnd, NULL, iCharWidth, iCharHeight);
			SetCaretPos (xCaret * iCharWidth, yCaret * iCharHeight);
			ShowCaret (hWnd);
			break;

		case WM_KILLFOCUS:

			HideCaret (hWnd);
			DestroyCaret();
			break;

		case WM_ERASEBKGND:

			return (TRUE);	// this avoids screen flashes during scrolls
			break;

		case WM_KEYDOWN:

			switch (wParam)
			{
				case VK_HOME:
						out_buf->data_byte = 0x1D;
						Signal_iRMX(hWnd);
						break;
				case VK_LEFT:
						out_buf->data_byte = 0x1F;
						Signal_iRMX(hWnd);
						break;
				case VK_RIGHT:
						out_buf->data_byte = 0x19;
						Signal_iRMX(hWnd);
						break;
				case VK_UP:
						out_buf->data_byte = 0x1E;
						Signal_iRMX(hWnd);
						break;
				case VK_DOWN:
						out_buf->data_byte = 0x1C;
						Signal_iRMX(hWnd);
						break;
				case VK_DELETE:
						out_buf->data_byte = 0x7F;
						Signal_iRMX(hWnd);
						break;
				case VK_PRIOR:		// page up
						out_buf->data_byte = 0x1E;	// up
						Signal_iRMX(hWnd);
						out_buf->data_byte = 0x1D;	// home
						Signal_iRMX(hWnd);
						break;
				case VK_NEXT:		// page down
						out_buf->data_byte = 0x1C;	// down
						Signal_iRMX(hWnd);
						out_buf->data_byte = 0x1D;	// home
						Signal_iRMX(hWnd);
						break;
			}
			break;

		case WM_CHAR:

			/*
			** if connected and iRMX buffer is empty,
			** then copy this keyboard characters to
			** iRMX input buffer
			*/

			if ((!(out_buf->flag_byte)) && (Connected))
			{
				KB_Char = (char) wParam;
				switch (KB_Char)
				{
					case '\x11':		// Cntrl-Q
						Stop_output = FALSE;
						out_buf->data_byte = KB_Char;
						break;
					case '\x13':		// Cntrl-S
						Stop_output = TRUE;
						out_buf->data_byte = KB_Char;
						break;
					case '\x1B':		// Escape
						out_buf->data_byte = 0x1B;
						break;
					case '\b':
						out_buf->data_byte = 0x7F;
						break;
					case '\t':
					case '\n':			// linefeed work ok
					case '\r':			// cariage return

					default:
						out_buf->data_byte = KB_Char;
						break;
				}
				Signal_iRMX(hWnd);
			}
			break;

		case WM_PAINT:

			Paint_Window (hWnd);
			break;

		case WM_COMMAND:			// message: command from application menu

			switch (wParam)
			{
				case IDM_ABOUT:

					lpProcAbout = MakeProcInstance (About, hInst);

					DialogBox (hInst,		// this instance
						"AboutBox",			// resource to use
						hWnd,				// parent handle
						lpProcAbout);		// About() instance address

					FreeProcInstance (lpProcAbout);
					break;

				case WTERM_EXIT:

					if (Connected)
					{
						Connected = FALSE;
						MessageBox (hWnd,
									"WTERM disconnected!",
									"WTERM",
									MB_ICONEXCLAMATION | MB_OK);
					}
					PostMessage (hWnd, WM_CLOSE, 0, 0L);
					break;

				case WTERM_CONNECT:

					if (!Connected)
					{
						if (Make_Connection (hWnd))
						{
							Connected = TRUE;
							ShowCaret (hWnd);
							SetWindowText (hWnd, (LPSTR)"WTERM - Connected");
							EnableMenuItem (GetMenu(hWnd), WTERM_CONNECT, MF_GRAYED);
							EnableMenuItem (GetMenu(hWnd), WTERM_SHUTDOWN, MF_ENABLED);
							/*
							** Force a character input so that the
							** iRMX Driver will echo something on 
							** connection
							*/
							out_buf->data_byte = '\n';
							Signal_iRMX (hWnd);
						}
					}
					else
						MessageBox (hWnd, "Connection established", "WTERM", MB_OK);
					break;

				case WTERM_SHUTDOWN:

					/*
					** clear the screen
					*/
					frg_color	= 0;	// Black
					bkg_color	= 0xf;	// White
					memset (pBuffer, ' ', iLineLength * iNumberLines);
					InvalidateRect (hWnd, NULL, TRUE);
					UpdateWindow (hWnd);
					SetCaretPos (0, 0);
					xCaret = yCaret = 0;
					HideCaret (hWnd);
					Connected = FALSE;
					SetWindowText (hWnd, (LPSTR)"WTERM");
					EnableMenuItem (GetMenu(hWnd), WTERM_CONNECT, MF_ENABLED);
					EnableMenuItem (GetMenu(hWnd), WTERM_SHUTDOWN, MF_GRAYED);
					break;

				default:			// Lets Windows process it
					return (DefWindowProc(hWnd, message, wParam, lParam));
			}
			break;

		case WM_DESTROY:			// message: window being destroyed

			PostQuitMessage(0);
			break;

		default:		// Pass unused message back to Windows

			return (DefWindowProc(hWnd, message, wParam, lParam));

	}		// end switch message

	return (NULL);
}

/*------------------------------------------------------------------------*
 *  Process_Text_buffer                                                   *
 *	                                                                      *
 *	input:		HWND hWnd; handle to the window                           *
 *	            WORD Num_Chars; number of characters received             *
 *	            LPSTR buffer; pointer to the buffer containing the chars  *
 *	output:     none                                                      *
 *	returns:    none                                                      *
 *	globals:    xCaret                                                    *
 *	            yCaret                                                    *
 *	            iLineLength                                               *
 *	            iNumberLines                                              *
 *	            iCharWidth                                                *
 *	            iCharHeight                                               *
 *	            pBuffer                                                   *
 *	            frg_color                                                 *
 *	            bkg_color                                                 *
 *	calls:      Scroll_Window_Up                                          *
 *	            Scroll_Window_Down                                        *
 *	            SetUpDC                                                   *
 *	            ResetDC                                                   *
 *	            Windows                                                   *
 *	abstract:   This is the main routine that processes the characters    *
 *	            received from the remote application to EMULATE an S120   *
 *	            type terminal OUTPUT. It does not invalidate the Client   *
 *	            Area if only a LINE's worth of text is changed - this     *
 *	            speeds up them display.                                   *
 *	                                                                      *
 *------------------------------------------------------------------------*/

void Process_Text_Buffer (HWND hWnd, WORD Num_Chars, LPSTR buffer)

{
HDC			hdc;
unsigned	i, k, temp;
RECT		rcInvalid;


hdc = GetDC(hWnd);
SetUpDC (hWnd, hdc);	//Setup the Device Context; colour, font, etc..

// scan the text, handle any special characters, and display the rest.

HideCaret (hWnd);
for (i=0; i < Num_Chars; i++)
{
	temp = (char) buffer[i];		// save for routine index case

	switch (Routine_index)
	{
	case 0 :	// routine$index = 0 => ordinary character
		{
			switch (buffer[i])
			{
			case 0 :			/* CASE 0 -	 */
					break;
			case 1 :			/* CASE 1 - erase rest of screen */

				/*
				** Fill the rest of the buffer with SPACEs
				*/
				k = (yCaret * iLineLength) + xCaret;
				memset ((pBuffer + k), ' ', (iLineLength*iNumberLines) -k);

				/*
				** now calculate the region to be invalid. We do this
				** so that at the next WM_PAINT message, only the 
				** invalid area needs to be redrawn.
				*/
				rcInvalid.top = yCaret;
				rcInvalid.bottom = iNumberLines;
				rcInvalid.left = 0;
				rcInvalid.right = iLineLength;
				LPtoDP(hdc,(POINT *) &rcInvalid, 2);

				// invalidate the area so that it gets redrawn
				
				InvalidateRect (hWnd, &rcInvalid, FALSE);
				UpdateWindow (hWnd);	// cause an immediate WM_PAINT

				break;

			case 2 :			/* CASE 2 - erase a WHOLE line */
				// Fill the Line with SPACEs
				k = (yCaret * iLineLength);
				memset ((pBuffer + k), ' ', iLineLength);

				// Output the Line using the current colors
				TextOut(hdc,
						0,
						yCaret,
						pBuffer + k,		// Clear from start of line
						iLineLength);		// SIZE of line

					break;

			case 3 :			/* CASE 3 - erase the REST of line */
				/*
				** Fill the rest of the line with SPACEs
				*/
				k = (yCaret * iLineLength) + xCaret;
				memset ((pBuffer + k), ' ', iLineLength - xCaret);
				TextOut(hdc,
						xCaret,
						yCaret,
						pBuffer + k,			// Clear line from X,Y
						iLineLength-xCaret);	// SIZE of rest of line

				break;

			case 4 :			/* CASE 4 - cursor absolute */

				Routine_index = 1;
				break;

			case 5 :			/* CASE 5 - line delete */

				Scroll_Window_Up (yCaret, 0, hWnd, hdc);
				break;

			case 6 :			/* CASE 6 - line insert */

				Scroll_Window_Down (yCaret, 0, hWnd, hdc);
				break;

			case 7 :			/* CASE 7 - bell */

				MessageBeep(2);
				break;

			case 8 :			/* CASE 8 - backspace */

				if(xCaret > 0)
					--xCaret;
				break;

			case 9 :			/* CASE 9 - tab */

				break;

			case 10 :			/* CASE 0A - linefeed */

				if (yCaret >= iNumberLines-1)
					Scroll_Window_Up (0, 0, hWnd, hdc);
				else
					++yCaret;
				break;

			case 11 :			/* CASE 0B - set color */

				Routine_index = 3;
				break;

			case 12 :			/* CASE 0C - erase screen */

				memset (pBuffer, ' ', iLineLength * iNumberLines);
					
				xCaret = 0;
				yCaret = 0;

				// invalidate the area so that it gets redrawn
				
				InvalidateRect (hWnd, NULL, FALSE);
				UpdateWindow (hWnd);

				break;

			case 13 :			/* CASE 0D - CR */

				xCaret = 0;		// move to beginning of the line
				break;

			case 14 :			/* CASE 0E - literal */
				break;
			case 15 :			/* CASE 0F - set raw mode */
				break;
			case 16 :			/* CASE 10 - set cooked mode */
				break;
			case 17 :			/* CASE 11 - X-ON	*/
				break;
			case 18 :			/* CASE 12		 */
				break;
			case 19 :			/* CASE 13 - X-OFF */
				break;
			case 20 :			/* CASE 14		 */
				break;
			case 21 :			/* CASE 15		 */
				break;
			case 22 :			/* CASE 16		 */
				break;
			case 23 :			/* CASE 17		 */
				break;
			case 24 :			/* CASE 18		 */
				break;

			case 25 :			/* CASE 19 - cursor right */

				if (xCaret >= iLineLength-1)
				{
					if (yCaret < iNumberLines-1)
					{
						++yCaret;
						xCaret = 0;
					}
				}
				else
					++xCaret;
				break;

			case 26 :			/* CASE 1A  */
					break;
			case 27 :			/* CASE 1B */
					break;

			case 28 :			/* CASE 1C - cursor down */

				if (yCaret != iNumberLines-1)
					++yCaret;
				break;

			case 29 :			/* CASE 1D - cursor home */

				xCaret = 0;
				yCaret = 0;
				break;

			case 30 :			/* CASE 1E - cursor up */

				if(yCaret != 0)
					--yCaret;
				break;

			case 31 :			/* CASE 1F - cursor left */

				if (xCaret == 0)
				{
					if (yCaret > 0)
					{
						--yCaret;
						xCaret = iLineLength-1;
					}
				}
				else
					--xCaret;
				break;

			default:		// all displayable characters

				// Filter out all characters not supported by windows

				if  ((buffer[i] < 32) ||
					((buffer[i] > 126) && (buffer[i] < 145)) ||
					((buffer[i] > 146) && (buffer[i] < 160)))
					break;

				// add char to screen buffer

				BUFFER (xCaret, yCaret) = (char) buffer[i];
				TextOut(hdc,
						xCaret,
						yCaret,
						&BUFFER (xCaret, yCaret),
						1);

				if (++xCaret > iLineLength-1)
				{
					xCaret = 0;
					if (yCaret >= iNumberLines-1)
						Scroll_Window_Up (0, 0, hWnd, hdc);
					else
						++yCaret;
				}
				break;

			}	/* CASE char */

		}	/* of case 0	=> routine$index = 0 */

		break;

	case 1 :	// routine$index = 1 - second byte of absolute cursor position
		{
			Routine_index = 2;
			temp -= 0x20;
			if (temp < iNumberLines)
				yCaret = temp;
			else
				yCaret = 0;
		}
		break;

	case 2 :	// routine$index = 2 - third byte of absolute cursor position
		{
			Routine_index = 0;
			temp -= 0x20;
			if (temp < iLineLength)
				xCaret = temp;
			else
				xCaret = 0;
		}
		break;
	case 3 :	// routine$index = 3 - second byte of set foreground color
		{
			// get foreground colour (bottom Nibble of the second byte)
			frg_color = (unsigned char) (temp & 0x0f);
			SetTextColor (hdc, GetNearestColor (hdc, color_tab[frg_color]));
			
			// get background colour (top Nibble of the second byte)
			bkg_color = (unsigned char) ((temp >> 4) & 0x0f);
			SetBkColor (hdc, GetNearestColor (hdc, color_tab[bkg_color]));
			
			Routine_index = 0;
		}
		break;

		default:
			break;

	}	// end of (switch (Routine_index))

	/*
	** set the cursor to the new position
	*/
	SetCaretPos (xCaret * iCharWidth, yCaret * iCharHeight);

}	// end	for(i=0; i < Num_Chars; i++)

ShowCaret (hWnd);
ResetDC (hdc);
ReleaseDC (hWnd, hdc);
}

/*------------------------------------------------------------------------*
 *  Scroll_Window_Down                                                    *
 *	                                                                      *
 *	input:		int row; row number from where to start scroll            *
 *	            int column; column number from where to start scroll (ign)*
 *	      		HWND hWnd; handle to the window                           *
 *	      		HDC  hDC;  handle to the Device Context                   *
 *	output:     none                                                      *
 *	returns:    none                                                      *
 *	globals:    iLineLength                                               *
 *	            iNumberLines                                              *
 *	            iCharHeight                                               *
 *	            pBuffer                                                   *
 *	calls:      Windows                                                   *
 *	abstract:   Scrolls the Client Area DOWN starting from the given row. *
 *	            Copies the line starting from the given row over the next *
 *	            line till the end of the SCREEN buffer (pseudo screen) and*
 *	            then calls Windows to do the actual scroll of the Client  *
 *	            Area. The uncovered area in invalidated and will get      *
 *	            painted at the next WM_PAINT msg. The given row is then   *
 *	            manually cleared.                                         *
 *	                                                                      *
 *------------------------------------------------------------------------*/

void Scroll_Window_Down (int row, int column, HWND hWnd, HDC hdc)
{
	int			y, z;
    RECT		rcRect;

	// scroll down one line starting from the given ROW

	for (z=(iNumberLines-1); z > row; z--)
	{
		/*
		** Copy a line OVER the NEXT line
		** y = previous line; z = current line
		*/
		y = z-1;
		memcpy ((pBuffer + (z*iLineLength)), (pBuffer + (y*iLineLength)), iLineLength);
	}

	/*
	** get the rectangle bounded by the
	** client area and change the TOP of
	** the rectangle to be the given ROW
	** then scroll this area only.
	*/
	GetClientRect (hWnd,&rcRect);
    DPtoLP (hdc, (POINT *) &rcRect, 2);
    rcRect.top = row;
	LPtoDP (hdc,(POINT *) &rcRect, 2);
    
	ScrollWindow (hWnd, 0, iCharHeight, &rcRect, &rcRect);

	/*
	** blank the given ROW
	*/
	y = row * iLineLength;
	memset ((pBuffer + y), ' ', iLineLength);
    
	UpdateWindow (hWnd);	//this causes a WM_PAINT immediately and
							// the RECT to be re-painted is the bit
							// uncovered AFTER the scroll and NOT
							// the WHOLE screen
}

/*------------------------------------------------------------------------*
 *  Scroll_Window_Up                                                      *
 *	                                                                      *
 *	input:		int row; row number from where to start scroll            *
 *	            int column; column number from where to start scroll (ign)*
 *	      		HWND hWnd; handle to the window                           *
 *	      		HDC  hDC;  handle to the Device Context                   *
 *	output:     none                                                      *
 *	returns:    none                                                      *
 *	globals:    iLineLength                                               *
 *	            iNumberLines                                              *
 *	            iCharHeight                                               *
 *	            pBuffer                                                   *
 *	calls:      Windows                                                   *
 *	abstract:   Scrolls the Client Area UP starting from the given row.   *
 *	            Copies the line starting from the given row over the prev.*
 *	            line till the top of the SCREEN buffer (pseudo screen) and*
 *	            then calls Windows to do the actual scroll of the Client  *
 *	            Area. The uncovered area in invalidated and will get      *
 *	            painted at the next WM_PAINT msg. The last line is then   *
 *	            manually cleared.                                         *
 *	                                                                      *
 *------------------------------------------------------------------------*/

void Scroll_Window_Up (int row, int column, HWND hWnd, HDC hdc)
{
	static		int	y, z;
    RECT		rcRect;

	// scroll up one line

	for (z=row; z < iNumberLines-1; z++)
	{
		/*
		** Copy a line OVER the PREVIOUS line
		** y = next line; z = current line
		*/
		y = z+1;
		memcpy ((pBuffer + (z*iLineLength)), (pBuffer + (y*iLineLength)), iLineLength);
	}

	/*
	** get the rectangle bounded by the
	** client area and set the TOP to the
	** given ROW and scroll this area only.
	*/
	GetClientRect (hWnd,&rcRect);
    DPtoLP (hdc, (POINT *) &rcRect, 2);
    rcRect.top = row;
	LPtoDP (hdc,(POINT *) &rcRect, 2);
    
	ScrollWindow (hWnd, 0, -(iCharHeight), &rcRect, &rcRect);
	
	/*
	** blank the last line
	*/
	y =  (iNumberLines-1) * iLineLength;
	memset ((pBuffer + y), ' ', iLineLength);

	UpdateWindow (hWnd);	//this causes a WM_PAINT immediately and
							// the RECT to be re-painted is the bit
							// uncovered AFTER the scroll and NOT
							// the WHOLE screen
}

/*------------------------------------------------------------------------*
 *  About                                                                 *
 *	                                                                      *
 *	input:		HWND hDlg; handle to the Dialog Box                       *
 *	            unsigned message; message ID                              *
 *	            WORD wParam; message parameter                            *
 *	            LONG lParam; message parameter                            *
 *	output:     none                                                      *
 *	returns:    none                                                      *
 *	globals:    none                                                      *
 *	calls:      Windows                                                   *
 *	abstract:                                                             *
 *	            Standard About Dialog Box window and exits when Left      *
 *	            mouse button or OK is pressed.                            *
 *	                                                                      *
 *------------------------------------------------------------------------*/

BOOL FAR PASCAL About (HANDLE   hDlg,
					   unsigned message,
					   WORD		wParam,
					   LONG		lParam)

{
	switch(message)					  // message: initialize dialog box
	{
		case WM_COMMAND:			  // message: received a command
		case WM_LBUTTONDOWN:

			EndDialog (hDlg, TRUE);
			return TRUE;

		case WM_INITDIALOG:
			return TRUE;

		default:					  // Didn't process a message
			return FALSE;
	}
}

/*------------------------------------------------------------------------*
 *  Signal_iRMX                                                           *
 *	                                                                      *
 *	input:		HWND hWnd; handle to the windows                          *
 *	output:     none                                                      *
 *	returns:    none                                                      *
 *	globals:    out_buf                                                   *
 *	calls:      Windows                                                   *
 *	            DOS Real-Time Extension to send a message to iRMX task    *
 *	abstract:                                                             *
 *	            Simply sets a flag in the shared data area to indicate    *
 *	            presence of key entry and then sends a message to the     *
 *	            iRMX Driver to tell is to pick it up.                     *
 *	                                                                      *
 *------------------------------------------------------------------------*/

void Signal_iRMX (HWND hWnd)
{
	out_buf->flag_byte = TRUE;

	// Send a message to iRMX driver

	RQSendMessage (Kb_Mux, segment, response, &except);
	if (except != EOK)
	{
		MessageBox (hWnd,
			"SendMessage error!",
			"WTERM",
			MB_ICONEXCLAMATION | MB_OK);

		// Add your own error handler here
	}
}

/*------------------------------------------------------------------------*
 *	SetupDC                                                               *
 *	                                                                      *
 *	input:		HWND hWnd; handle to the window                           *
 *	            HDC hdc; handle to the Device Context                     *
 *	output:     none                                                      *
 *	returns:    none                                                      *
 *	globals:    color_tab                                                 *
 *	            frg_color                                                 *
 *	            bkg_color                                                 *
 *	calls:      Windows                                                   *
 *	abstract:   Sets up the mapping mode of the Client area to map logical*
 *	            coordinates to pixels where the logical coordinates are   *
 *	            character position on the Client Area.                    *
 *	            Also sets up the current foreground and background colours*
 *	                                                                      *
 *------------------------------------------------------------------------*/

void SetUpDC (HWND hWnd, HDC hDC)
{
    RECT rClRect;

    GetClientRect (hWnd, &rClRect);

    // set origin of the window
	SetViewportOrg (hDC, 0, 0);

    SetMapMode (hDC, MM_ANISOTROPIC);	// Logical units mapped to arbitrarily
    									//scaled X, Y axes

    // Set the extents such that one unit horizontally or 
    // vertically is one character width or height.
    SetWindowExt (hDC, 1, 1);
    SetViewportExt (hDC, iCharWidth, iCharHeight);

    // Set the background mode to opaque, and select the SYSTEM font.
    SetBkMode (hDC, OPAQUE);
    hOldFont = SelectObject (hDC, GetStockObject (SYSTEM_FIXED_FONT));

	SetTextColor (hDC, GetNearestColor (hDC, color_tab[frg_color]));
	SetBkColor (hDC, GetNearestColor (hDC, color_tab[bkg_color]));
}

/*------------------------------------------------------------------------*
 *	ResetDC                                                               *
 *	                                                                      *
 *	input:		HDC hdc; handle to the Device Context                     *
 *	output:     none                                                      *
 *	returns:    none                                                      *
 *	globals:    none                                                      *
 *	calls:      Windows                                                   *
 *	abstract:   Reselects the old font currently                          *
 *	                                                                      *
 *------------------------------------------------------------------------*/

void ResetDC (HDC hDC)
{
    SelectObject (hDC, hOldFont);
}

/*------------------------------------------------------------------------*
 *	Paint_window                                                          *
 *	                                                                      *
 *	input:		HWND hWnd; handle to the window                           *
 *	output:     none                                                      *
 *	returns:    none                                                      *
 *	globals:    none                                                      *
 *	calls:      Windows                                                   *
 *	abstract:   Redraws the invalidated Client Area. It redraws it bit    *
 *	            more than what Windows tells us to get around some obscure*
 *	            BUG in Windows that is seen when an ICON overlaid on top  *
 *	            of our Client Area while the Area is scrolling. We expect *
 *	            the invalid area the encompass ALL of the ICON but Windows*
 *	            only seem to give part of it.                             *
 *	                                                                      *
 *------------------------------------------------------------------------*/

void Paint_Window (HWND hWnd)
{
    register	int i;
    PAINTSTRUCT	ps;
    HDC			hDC;
    RECT		rcUpdate;
    int			nVPaintBeg, nVPaintEnd, nHPaintBeg, nHPaintEnd;

    hDC = BeginPaint (hWnd, (LPPAINTSTRUCT) &ps);
    SetUpDC (hWnd, hDC);

    rcUpdate = ps.rcPaint;
    DPtoLP (hDC, (POINT *) &rcUpdate, 2);

    // calculate first and last lines to update
    nVPaintBeg = max (0, rcUpdate.top-2);
    nVPaintEnd = min (iNumberLines, rcUpdate.bottom+1);

    // calculate the first and last columns to update
    nHPaintBeg = max (0, rcUpdate.left-1);
    nHPaintEnd = min (iLineLength, rcUpdate.right+1);

    // display the lines that need to be drawn
    for (i = nVPaintBeg; i < nVPaintEnd; i++)
    {
        TextOut (hDC,
				 nHPaintBeg,
                 i,
				 pBuffer + (i * iLineLength) + nHPaintBeg,
				 iLineLength-nHPaintBeg);
    }
    ResetDC (hDC);
    EndPaint (hWnd, (LPPAINTSTRUCT)&ps);
}

/*------------------------------------------------------------------------*
 *	Make_Connection                                                       *
 *	                                                                      *
 *	input:		HDC hWnd-handle to the window                             *
 *	output:     none                                                      *
 *	returns:    none                                                      *
 *	globals:    none                                                      *
 *	calls:      Windows                                                   *
 *	abstract:   Checks the presence of iRMX.                              *
 *	                                                                      *
 *------------------------------------------------------------------------*/

BOOL Make_Connection (HWND hWnd)
{
	/*
	** Make some calls to the RT extension to find out if the RMX
	** keyboard mail box is present.
	** NOTE:
	**		If the RT extension is not configured then
	**		Windows (and DOS) will get confused!!!
	*/

	roottoken = RQGetTaskTokens (rootjob,&except);

	// Display debug messages - you can delete the following four lines
	// and supply your own error handler
	if (except != EOK)
	{
		MessageBox (hWnd, "gettasktoken error", "WTERM", MB_OK);
		return(FALSE);
	}

	vm86j = RQLookupObject (roottoken,&vm86[0],0,&except);

	// Display debug messages - you can delete the following four lines
	// and supply your own error handler
	if (except != EOK)
	{
		MessageBox(hWnd, "lookupobject root error", "WTERM", MB_OK);
		return(FALSE);
	}

	Kb_Mux = RQLookupObject (vm86j,&kb_mb[0],0,&except);
	segment	= Kb_Mux;

	// Display debug messages - you can delete the following four lines
	// and supply your own error handler
	if (except != EOK)
	{
		MessageBox(hWnd, "lookupobject vm86 error", "WTERM", MB_OK);
		return(FALSE);
	}
	return(TRUE);
}

//*************************  End Of Source Code Module *********************
